from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory, EveCharacterFactory
from app_utils.testing import create_state

from charlink.utils import get_visible_corps, chars_annotate_linked_apps, get_user_available_apps, get_user_linked_chars
from charlink.app_imports import import_apps
from charlink.imports.corptools import _corp_perms


class TestGetVisibleCorps(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.superuser = UserMainFactory(is_superuser=True)

        cls.main_char = cls.user.profile.main_character

        cls.corporation = cls.user.profile.main_character.corporation
        cls.alliance = cls.corporation.alliance
        cls.state = cls.user.profile.state

        cls.corporation2 = EveCorporationInfoFactory(create_alliance=False)
        cls.corporation2.alliance = cls.alliance
        cls.corporation2.save()
        char = EveCharacterFactory(corporation=cls.corporation2)
        UserMainFactory(main_character__character=char)

        cls.corporation3 = EveCorporationInfoFactory()
        cls.alliance2 = cls.corporation3.alliance
        cls.state.member_alliances.add(cls.alliance2)
        char = EveCharacterFactory(corporation=cls.corporation3)
        UserMainFactory(main_character__character=char)

        cls.corporation_empty = EveCorporationInfoFactory(alliance=cls.alliance2)

        cls.corporation_superuser = cls.superuser.profile.main_character.corporation
        cls.alliance_superuser = cls.corporation2.alliance
        cls.state_superuser = create_state(1000, member_alliances=[cls.alliance_superuser])

    def test_superuser(self):
        corps = get_visible_corps(self.superuser)
        self.assertQuerysetEqual(
            corps,
            [
                self.corporation,
                self.corporation_superuser,
                self.corporation2,
                self.corporation3,
            ],
            ordered=False
        )

    def test_corp_access(self):
        AuthUtils.add_permission_to_user_by_name('charlink.view_corp', self.user)
        corps = get_visible_corps(self.user)
        self.assertQuerysetEqual(
            corps,
            [
                self.corporation,
            ],
            ordered=False
        )

    def test_alliance_access(self):
        AuthUtils.add_permission_to_user_by_name('charlink.view_alliance', self.user)
        corps = get_visible_corps(self.user)
        self.assertQuerysetEqual(
            corps,
            [
                self.corporation,
                self.corporation2,
            ],
            ordered=False
        )

    def test_state_access(self):
        AuthUtils.add_permission_to_user_by_name('charlink.view_state', self.user)
        corps = get_visible_corps(self.user)
        self.assertQuerysetEqual(
            corps,
            [
                self.corporation3,
            ],
            ordered=False
        )

    def test_state_and_alliance_access(self):
        AuthUtils.add_permissions_to_user_by_name(['charlink.view_state', 'charlink.view_alliance'], self.user)
        corps = get_visible_corps(self.user)
        self.assertQuerysetEqual(
            corps,
            [
                self.corporation,
                self.corporation2,
                self.corporation3,
            ],
            ordered=False
        )

    def test_no_access(self):
        corps = get_visible_corps(self.user)
        self.assertQuerysetEqual(
            corps,
            [],
            ordered=False
        )


class TestCharsAnnotateLinkedApps(TestCase):

    @classmethod
    def setUpTestData(cls):
        EveCharacterFactory.create_batch(10)

    def test_ok(self):
        chars = EveCharacter.objects.all()
        imported_apps = import_apps()

        res = chars_annotate_linked_apps(chars, [imported_apps['add_character'].imports[0]])

        self.assertEqual(len(res), 10)
        for char in res:
            self.assertTrue(hasattr(char, 'add_character_default'))


class TestGetUserAvailableApps(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["memberaudit.basic_access"])
        cls.corptools_user_corp = UserMainFactory(permissions=_corp_perms)
        cls.corptools_user_charaudit = UserMainFactory(permissions=["corptools.view_characteraudit"])

    def test_ok(self):
        res = get_user_available_apps(self.user)
        self.assertSetEqual(
            set(res.keys()),
            {'memberaudit', 'add_character'}
        )

        res2 = get_user_available_apps(self.corptools_user_corp)
        self.assertSetEqual(
            set(res2.keys()),
            {'add_character', 'corptools'}
        )

        res3 = get_user_available_apps(self.corptools_user_charaudit)
        self.assertSetEqual(
            set(res3.keys()),
            {'add_character', 'corptools'}
        )


class TestGetUserLinkedChars(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_ok(self):
        res = get_user_linked_chars(self.user)

        self.assertIn('apps', res)
        self.assertIn('characters', res)



# from django.test import TestCase
# from django.urls import reverse

# from app_utils.testdata_factories import UserMainFactory


# class TestHooks(TestCase):

#     @classmethod
#     def setUpTestData(cls):
#         cls.testuser = UserMainFactory()
#         cls.html_menu = f"""
#             <li>
#                 <a class="active" href="{reverse('charlink:index')}">
#                     <i class="fas fa-link"></i> CharLink
#                 </a>
#             </li>
#         """

#     def test_render_hook(self):
#         self.client.force_login(self.testuser)

#         response = self.client.get(reverse('charlink:index'))
#         self.assertContains(response, self.html_menu, html=True)



from unittest.mock import patch

from django.test import TestCase, RequestFactory

from app_utils.testdata_factories import UserMainFactory

from charlink.imports.corptools import _add_character_charaudit, _is_character_added_charaudit, _corp_perms, _is_character_added_corp, _add_character_corp
from charlink.app_imports import import_apps


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['corptools.view_characteraudit', *_corp_perms])
        cls.factory = RequestFactory()

    @patch('charlink.imports.corptools.update_character.apply_async')
    def test_ok_charaudit(self, mock_update_character):
        mock_update_character.return_value = None

        request = self.factory.get('/charlink/login/')
        token = self.user.token_set.first()

        _add_character_charaudit(request, token)

        mock_update_character.assert_called_once_with(args=[token.character_id], priority=6)
        self.assertTrue(_is_character_added_charaudit(self.user.profile.main_character))

    @patch('charlink.imports.corptools.update_all_corps.apply_async')
    def test_ok_corp(self, mock_update_all_corps):
        mock_update_all_corps.return_value = None

        request = self.factory.get('/charlink/login/')
        token = self.user.token_set.first()

        _add_character_corp(request, token)

        self.assertTrue(_is_character_added_corp(self.user.profile.main_character))
        self.assertTrue(mock_update_all_corps.called)


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.character = cls.user.profile.main_character
        cls.factory = RequestFactory()

    @patch('charlink.imports.corptools.update_character.apply_async')
    def test_ok_charaudit(self, mock_update_character):
        mock_update_character.return_value = None

        self.assertFalse(_is_character_added_charaudit(self.character))
        _add_character_charaudit(self.factory.get('/charlink/login/'), self.user.token_set.first())
        self.assertTrue(_is_character_added_charaudit(self.character))

    @patch('charlink.imports.corptools.update_all_corps.apply_async')
    def test_ok_corp(self, mock_update_all_corps):
        mock_update_all_corps.return_value = None

        self.assertFalse(_is_character_added_corp(self.character))
        _add_character_corp(self.factory.get('/charlink/login/'), self.user.token_set.first())
        self.assertTrue(_is_character_added_corp(self.character))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.charaudit_user = UserMainFactory(permissions=['corptools.view_characteraudit'])
        cls.corp_user = UserMainFactory(permissions=_corp_perms)

    def test_ok_charaudit(self):
        login_import = import_apps()['corptools'].get('default')

        self.assertFalse(login_import.check_permissions(self.no_perm_user))
        self.assertTrue(login_import.check_permissions(self.charaudit_user))

    def test_ok_corp(self):
        login_import = import_apps()['corptools'].get('structures')

        self.assertFalse(login_import.check_permissions(self.no_perm_user))
        self.assertTrue(login_import.check_permissions(self.corp_user))
        self.assertFalse(login_import.check_permissions(self.charaudit_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        UserMainFactory.create_batch(4)
        UserMainFactory.create_batch(3, permissions=['corptools.view_characteraudit'])
        UserMainFactory.create_batch(5, permissions=_corp_perms)

    def test_ok_charaudit(self):
        login_import = import_apps()['corptools'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 3)

    def test_ok_corp(self):
        login_import = import_apps()['corptools'].get('structures')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 5)



from django import forms

from .app_imports import import_apps
from .app_settings import CHARLINK_IGNORE_APPS


class LinkForm(forms.Form):
    add_character_default = forms.BooleanField(
        required=False,
        initial=True,
        disabled=True,
        label='Add Character (default)'
    )

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for app, imports in import_apps().items():
            if app != 'add_character' and app not in CHARLINK_IGNORE_APPS:
                self.fields.update(imports.get_form_fields(user))



from django.urls import path

from . import views

app_name = 'charlink'

urlpatterns = [
    path('', views.index, name='index'),
    path('login/', views.login_view, name='login'),
    path('audit/corp/<int:corp_id>/', views.audit, name='audit_corp'),
    path('audit/user/<int:user_id>/', views.audit_user, name='audit_user'),
    path('audit/app/<str:app>/', views.audit_app, name='audit_app'),
    path('search/', views.search, name='search'),
]



from django.test import TestCase
from django.template import Context, Template

from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory


class TestGetCorpMembersFilter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.testuser = UserMainFactory()
        cls.testcorp = cls.testuser.profile.main_character.corporation

        cls.corpmates = EveCharacterFactory.create_batch(5, corporation=cls.testcorp) + [cls.testuser.profile.main_character]

        cls.template = Template('{% load charlinkutils %}{% for char in corp|get_corp_members %}{{ char.character_name }},{% endfor %}')

    def test_success(self):
        context = Context({'corp': self.testcorp})

        res = self.template.render(context)

        for char in self.corpmates:
            self.assertIn(char.character_name, res)


class TestGetCharAttrFilter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.testuser = UserMainFactory()
        cls.testcharacter = cls.testuser.profile.main_character

        cls.template = Template('{% load charlinkutils %}{{ char|get_char_attr:"character_name" }}')

    def test_character_success(self):
        context = Context({'char': self.testcharacter})

        res = self.template.render(context)

        self.assertEqual(res, self.testcharacter.character_name)

    def test_int_success(self):
        context = Context({'char': self.testcharacter.pk})

        res = self.template.render(context)

        self.assertEqual(res, self.testcharacter.character_name)

    def test_int_fail(self):
        context = Context({'char': self.testcharacter.pk + 10})

        res = self.template.render(context)

        self.assertEqual(res, '')

    def test_str_fail(self):
        context = Context({'char': 'notanumber'})

        res = self.template.render(context)

        self.assertEqual(res, '')

    def test_str_success(self):
        context = Context({'char': str(self.testcharacter.pk)})

        res = self.template.render(context)

        self.assertEqual(res, self.testcharacter.character_name)

    def test_param_not_valid(self):
        context = Context({'char': []})

        res = self.template.render(context)

        self.assertEqual(res, '')



from django.test import TestCase

from charlink.app_imports import import_apps

from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory
from app_utils.testing import add_character_to_user


_scopes_readfleet = ["esi-fleets.read_fleet.v1"]
_scopes_clickfleet = ["esi-location.read_location.v1", "esi-location.read_ship_type.v1", "esi-location.read_online.v1"]


class TestAddCharacter(TestCase):

    def test_ok(self):
        app_import = import_apps()['afat']

        self.assertIsNone(app_import.get('readfleet').add_character(None, None))
        self.assertIsNone(app_import.get('clickfat').add_character(None, None))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.main_character = cls.user.profile.main_character

        cls.char_readfleet = EveCharacterFactory()
        add_character_to_user(cls.user, cls.char_readfleet, scopes=_scopes_readfleet)
        cls.char_clickfat = EveCharacterFactory()
        add_character_to_user(cls.user, cls.char_clickfat, scopes=_scopes_clickfleet)

    def test_ok(self):
        app_import = import_apps()['afat']

        self.assertFalse(app_import.get('readfleet').is_character_added(self.main_character))
        self.assertTrue(app_import.get('readfleet').is_character_added(self.char_readfleet))
        self.assertFalse(app_import.get('readfleet').is_character_added(self.char_clickfat))

        self.assertFalse(app_import.get('clickfat').is_character_added(self.main_character))
        self.assertFalse(app_import.get('clickfat').is_character_added(self.char_readfleet))
        self.assertTrue(app_import.get('clickfat').is_character_added(self.char_clickfat))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.manage_user = UserMainFactory(permissions=['afat.manage_afat'])
        cls.add_fatlink_user = UserMainFactory(permissions=['afat.add_fatlink'])
        cls.both_user = UserMainFactory(permissions=['afat.manage_afat', 'afat.add_fatlink'])
        cls.basic_access_user = UserMainFactory(permissions=['afat.basic_access'])
        cls.no_perm_user = UserMainFactory()

    def test_ok(self):
        app_import = import_apps()['afat']

        self.assertTrue(app_import.get('readfleet').check_permissions(self.manage_user))
        self.assertTrue(app_import.get('readfleet').check_permissions(self.add_fatlink_user))
        self.assertTrue(app_import.get('readfleet').check_permissions(self.both_user))
        self.assertFalse(app_import.get('readfleet').check_permissions(self.no_perm_user))

        self.assertFalse(app_import.get('clickfat').check_permissions(self.no_perm_user))
        self.assertTrue(app_import.get('clickfat').check_permissions(self.basic_access_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.manage_user = UserMainFactory(permissions=['afat.manage_afat'])
        cls.add_fatlink_user = UserMainFactory(permissions=['afat.add_fatlink'])
        cls.both_user = UserMainFactory(permissions=['afat.manage_afat', 'afat.add_fatlink'])
        cls.basic_access_user = UserMainFactory(permissions=['afat.basic_access'])
        cls.no_perm_user = UserMainFactory()

    def test_ok(self):
        app_import = import_apps()['afat']

        users_readfleet = app_import.get('readfleet').get_users_with_perms()
        self.assertQuerysetEqual(
            users_readfleet,
            [
                self.manage_user.pk,
                self.add_fatlink_user.pk,
                self.both_user.pk
            ],
            ordered=False,
            transform=lambda x: x.pk
        )

        users_clickfat = app_import.get('clickfat').get_users_with_perms()
        self.assertQuerysetEqual(
            users_clickfat,
            [
                self.basic_access_user.pk
            ],
            transform=lambda x: x.pk
        )



from importlib import import_module

from django.conf import settings
from django.contrib.auth.models import User
from django.db.models import Exists, OuterRef

from allianceauth.services.hooks import get_extension_logger
from allianceauth.authentication.models import CharacterOwnership

from .utils import LoginImport, AppImport

logger = get_extension_logger(__name__)

_supported_apps = {
    'add_character': AppImport('add_character', [
        LoginImport(
            app_label='add_character',
            unique_id='default',
            field_label='Add Character (default)',
            add_character=lambda request, token: None,
            scopes=['publicData'],
            check_permissions=lambda user: True,
            is_character_added=lambda character: CharacterOwnership.objects.filter(character=character).exists(),
            is_character_added_annotation=Exists(CharacterOwnership.objects.filter(character_id=OuterRef('pk'))),
            get_users_with_perms=lambda: User.objects.filter(
                Exists(CharacterOwnership.objects.filter(user_id=OuterRef('pk')))
            ),
        )
    ])
}

_imported = False


def import_apps():
    global _imported
    if not _imported:
        for app in settings.INSTALLED_APPS:
            if app != 'allianceauth':
                try:
                    module = import_module(f'charlink.imports.{app}')
                except ModuleNotFoundError:
                    logger.debug(f"Loading of {app} link: failed")
                else:
                    _supported_apps[app] = module.import_app
                    logger.debug(f"Loading of {app} link: success")

        _imported = True

    return _supported_apps



from allianceauth import hooks
from allianceauth.services.hooks import UrlHook, MenuItemHook

from . import urls


class CharlinkMenuItemHook(MenuItemHook):
    def __init__(self):
        super().__init__("CharLink", "fas fa-link", "charlink:index", navactive=['charlink:'])


@hooks.register('menu_item_hook')
def register_menu():
    return CharlinkMenuItemHook()


@hooks.register('url_hook')
def register_urls():
    return UrlHook(urls, 'charlink', 'charlink/')



from unittest.mock import patch, Mock

from django.test import TestCase
from django.urls import reverse
from django.contrib.messages import get_messages, DEFAULT_LEVELS
from django.db.models import OuterRef, Exists

from allianceauth.authentication.models import CharacterOwnership

from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory, EveCharacterFactory

from charlink.views import get_navbar_elements
from charlink.imports.memberaudit import import_app as memberaudit_import
from charlink.imports.miningtaxes import import_app as miningtaxes_import
from charlink.imports.corptools import _corp_perms
from charlink.app_imports.utils import AppImport, LoginImport


class TestGetNavbarElements(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.permuser = UserMainFactory(permissions=['charlink.view_corp'])
        cls.nopermuser = UserMainFactory()

    def test_with_perm(self):
        res = get_navbar_elements(self.permuser)

        self.assertIn('available', res)
        self.assertTrue(res['is_auditor'])

        self.assertIn('available', res)
        self.assertGreater(len(res['available']), 0)

        self.assertIn('available_apps', res)
        self.assertGreater(len(res['available_apps']), 0)

    def test_without_perm(self):
        res = get_navbar_elements(self.nopermuser)

        self.assertIn('available', res)
        self.assertFalse(res['is_auditor'])

        self.assertIn('available', res)
        self.assertEqual(len(res['available']), 0)

        self.assertIn('available_apps', res)
        self.assertEqual(len(res['available_apps']), 0)


class TestIndex(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=[
            'corputils.add_corpstats',
            'memberaudit.basic_access',
            'miningtaxes.basic_access',
            'moonmining.add_refinery_owner',
            'moonmining.basic_access',
            'corpstats.add_corpstat',
        ])

        cls.form_data = {
            'allianceauth.corputils_default': ['on'],
            'miningtaxes_default': ['on'],
            'moonmining_default': ['on'],
            'corpstats_default': ['on'],
        }

    def test_get(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:index'))

        self.assertEqual(res.status_code, 200)
        self.assertIn('form', res.context)
        self.assertIn('characters_added', res.context)

    def test_post_ok(self):
        self.client.force_login(self.user)

        res = self.client.post(reverse('charlink:index'), self.form_data)

        self.assertRedirects(
            res,
            reverse('charlink:login'),
            fetch_redirect_response=False
        )

        session = self.client.session

        self.assertIn('charlink', session)
        self.assertIn('scopes', session['charlink'])
        self.assertIn('imports', session['charlink'])

        converted_imports = [list(x) for x in session['charlink']['imports']]

        self.assertIn(['add_character', 'default'], converted_imports)
        self.assertIn(['allianceauth.corputils', 'default'], converted_imports)
        self.assertNotIn(['memberaudit', 'default'], converted_imports)
        self.assertIn(['miningtaxes', 'default'], converted_imports)
        self.assertIn(['moonmining', 'default'], converted_imports)
        self.assertIn(['corpstats', 'default'], converted_imports)
        self.assertEqual(len(session['charlink']['imports']), 5)

    # form always valid
    # def test_post_wrong_data(self):
    #     self.client.force_login(self.user)

    #     res = self.client.post(reverse('charlink:index'), {'add_character:': '5'})

    #     self.assertEqual(res.status_code, 200)
    #     self.assertIn('form', res.context)
    #     self.assertIn('characters_added', res.context)

    # force form invalid
    @patch('charlink.forms.LinkForm.is_valid')
    def test_form_invalid(self, mock_is_valid):
        mock_is_valid.return_value = False

        self.client.force_login(self.user)
        res = self.client.post(reverse('charlink:index'), self.form_data)

        self.assertEqual(res.status_code, 200)
        self.assertIn('form', res.context)
        self.assertIn('characters_added', res.context)


class TestLoginView(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.scopes = list(set(memberaudit_import.imports[0].scopes + miningtaxes_import.imports[0].scopes))
        cls.user = UserMainFactory(
            permissions=[
                'memberaudit.basic_access',
                'miningtaxes.basic_access',
            ],
            main_character__scopes=cls.scopes
        )
        cls.token = cls.user.token_set.first()

    @patch('charlink.imports.miningtaxes._add_character')
    @patch('charlink.imports.memberaudit._add_character')
    @patch('charlink.views.import_apps')
    @patch('charlink.decorators.token_required')
    def test_ok(self, mock_token_required, mock_import_apps, mock_memberaudit_add_character, mock_miningtaxes_add_character):
        session = self.client.session
        session['charlink'] = {
            'scopes': self.scopes,
            'imports': [
                ('memberaudit', 'default'),
                ('miningtaxes', 'default'),
                ('add_character', 'default'),
            ],
        }
        session.save()

        def fake_decorator(f):
            def fake_wrapper(request, *args, **kwargs):
                return f(request, self.token, *args, **kwargs)
            return fake_wrapper

        mock_token_required.return_value = fake_decorator

        mock_import_apps.return_value = {
            'memberaudit': AppImport('memberaudit', [
                LoginImport(
                    app_label='memberaudit',
                    unique_id='default',
                    field_label='Member Audit',
                    add_character=lambda request, token: None,
                    scopes=memberaudit_import.imports[0].scopes,
                    check_permissions=memberaudit_import.imports[0].check_permissions,
                    is_character_added=memberaudit_import.imports[0].is_character_added,
                    is_character_added_annotation=memberaudit_import.imports[0].is_character_added_annotation,
                    get_users_with_perms=memberaudit_import.imports[0].get_users_with_perms,
                )
            ]),
            'miningtaxes': AppImport('miningtaxes', [
                LoginImport(
                    app_label='miningtaxes',
                    unique_id='default',
                    field_label='Mining Taxes',
                    add_character=Mock(side_effect=Exception('test')),
                    scopes=miningtaxes_import.imports[0].scopes,
                    check_permissions=miningtaxes_import.imports[0].check_permissions,
                    is_character_added=miningtaxes_import.imports[0].is_character_added,
                    is_character_added_annotation=miningtaxes_import.imports[0].is_character_added_annotation,
                    get_users_with_perms=miningtaxes_import.imports[0].get_users_with_perms,
                )
            ]),
            'add_character': AppImport('add_character', [
                LoginImport(
                    app_label='add_character',
                    unique_id='default',
                    field_label='Add Character (default)',
                    add_character=lambda request, token: None,
                    scopes=['publicData'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: CharacterOwnership.objects.filter(character=character).exists(),
                    is_character_added_annotation=Exists(CharacterOwnership.objects.filter(character_id=OuterRef('pk'))),
                    get_users_with_perms=lambda: None,
                )
            ]),
        }

        mock_memberaudit_add_character.return_value = None
        mock_miningtaxes_add_character.side_effect = Exception('test')

        self.client.force_login(self.user)
        res = self.client.get(reverse('charlink:login'))

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 2)

        sorted_messages = sorted(messages, key=lambda x: x.level)
        self.assertEqual(sorted_messages[0].level, DEFAULT_LEVELS['SUCCESS'])
        self.assertEqual(sorted_messages[1].level, DEFAULT_LEVELS['ERROR'])


class TestAudit(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_corp'])
        cls.corp = cls.user.profile.main_character.corporation
        cls.corp2 = EveCorporationInfoFactory()
        cls.char2 = EveCharacterFactory(corporation=cls.corp2)
        cls.user2 = UserMainFactory(main_character__character=cls.char2)

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_corp', args=[self.corp.corporation_id]))

        self.assertEqual(res.status_code, 200)
        self.assertIn('selected', res.context)

    def test_no_perm(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_corp', args=[self.corp2.corporation_id]))

        self.assertNotEqual(res.status_code, 200)


class TestSearch(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_corp'])
        cls.main_char = cls.user.profile.main_character

        cls.user2 = UserMainFactory()
        cls.main_char2 = cls.user2.profile.main_character

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:search'), {'search_string': self.main_char.character_name})

        self.assertEqual(res.status_code, 200)
        self.assertIn('search_string', res.context)
        self.assertIn('characters', res.context)
        self.assertEqual(len(res.context['characters']), 1)

    def test_not_found(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:search'), {'search_string': self.main_char2.character_name})

        self.assertEqual(res.status_code, 200)
        self.assertIn('search_string', res.context)
        self.assertIn('characters', res.context)
        self.assertEqual(len(res.context['characters']), 0)

    def test_missing_string(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:search'))

        self.assertRedirects(res, reverse('charlink:index'))


class TestAuditUser(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_corp'])

        char2 = EveCharacterFactory(corporation=cls.user.profile.main_character.corporation)
        cls.user2 = UserMainFactory(main_character__character=char2)

        cls.user_ext = UserMainFactory()

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_user', args=[self.user2.pk]))

        self.assertEqual(res.status_code, 200)
        self.assertIn('characters_added', res.context)

    def test_no_perm(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_user', args=[self.user_ext.pk]))

        self.assertNotEqual(res.status_code, 200)


class TestAuditApp(TestCase):

    @classmethod
    def setUpTestData(cls):
        permissions = ["memberaudit.basic_access", "moonmining.add_refinery_owner", "moonmining.basic_access"]
        cls.user = UserMainFactory(permissions=[
            'charlink.view_corp',
            'corptools.view_characteraudit',
            *permissions,
            *_corp_perms,
        ])
        char2, char3 = EveCharacterFactory.create_batch(
            2,
            corporation=cls.user.profile.main_character.corporation
        )
        cls.user2 = UserMainFactory(
            permissions=permissions,
            main_character__character=char2
        )
        cls.random_char = EveCharacterFactory(corporation=cls.user.profile.main_character.corporation)
        cls.no_perm_user = UserMainFactory(
            permissions=['charlink.view_corp'],
            main_character__character=char3
        )

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['memberaudit']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('app', res.context)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 1)
        self.assertEqual(len(list(res.context['logins'].values())[0]), 2)

    def test_app_empty_perms(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['add_character']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 1)
        self.assertEqual(len(list(res.context['logins'].values())[0]), 3)
        self.assertIn('app', res.context)

    def test_missing_app(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['invalid_app']))

        self.assertEqual(res.status_code, 404)

    def test_no_app_perm(self):
        self.client.force_login(self.no_perm_user)

        res = self.client.get(reverse('charlink:audit_app', args=['memberaudit']))

        self.assertNotEqual(res.status_code, 200)

    def test_multiple_app_perms(self):
        self.client.force_login(self.user)

        extra_char = EveCharacterFactory(corporation=self.user.profile.main_character.corporation)
        UserMainFactory(
            permissions=["moonmining.add_refinery_owner", "moonmining.basic_access"],
            main_character__character=extra_char
        )

        res = self.client.get(reverse('charlink:audit_app', args=['moonmining']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 1)
        self.assertEqual(len(list(res.context['logins'].values())[0]), 3)
        self.assertIn('app', res.context)

    def test_app_with_multiple_logins(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['corptools']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 2)
